/*
     This program synchronizes the simple producer-consumer
     example in the program 1-trivialPC-unsynchronized.c.

     What we have is a producer-consumer example with a bounded
     buffer of size one (not a very big buffer). This program uses
     two binary semaphores to do the synchronization by "bounding"
     the buffer.

     With the semaphores, the "producer" cannot increment the
     counter until the "consumer" has printed out the current
     value of the counter. And the "consumer" cannot print out
     another value until the "producer" has done another increment.
     So the output of this program is all of the integers in
     consecutive order.

     Notice that the problem in 1-trivialPC-unsynchronized.c is
     not a critical section problem. We cannot solve the problem in
     1-trivialPC-unsynchronized.c by using a mutex. The synchronization
     done below with the semaphores is a different synchronization
     pattern than the "mutual exclusion" needed for critical sections.
*/
#include <windows.h>
#include <stdio.h>

// global variables
static int runFlag = TRUE;
static int count = 0;

HANDLE semaphoreI;  // the incremented semaphore
HANDLE semaphoreP;  // the printed semaphore

int main( int argc, char *argv[])
{
   //thread function prototypes
   DWORD WINAPI thread1(LPVOID);
   DWORD WINAPI thread2(LPVOID);

   DWORD thread1ID;
   DWORD thread2ID;
   int threadParam1 = 0;  // thread function parameter
   int threadParam2 = 0;

   // create two Semaphore objects (one signaled, one not signaled)
   semaphoreP = CreateSemaphore(NULL, 1, 1, NULL); // signaled
   semaphoreI = CreateSemaphore(NULL, 0, 1, NULL); // not signaled

   // create threads
   printf("Starting threads\n");
   CreateThread(NULL, 0, thread1, &threadParam1, 0, &thread1ID);
   CreateThread(NULL, 0, thread2, &threadParam2, 0, &thread2ID);

   printf("The incrementer is thread number %d.\n", thread1ID);
   printf("The printer is thread number %d.\n", thread2ID);

   Sleep(800);     // let the threads run for a while
   runFlag = FALSE; // now stop the threads
   Sleep(200);      // give them time to stop
   printf("\nThreads stopped.\n");
   printf("count = %d\n", count);

   return 0;
}//main


DWORD WINAPI thread1(LPVOID threadParam)
{
   while(runFlag)
   {
      WaitForSingleObject(semaphoreP, INFINITE);
      count++;
      ReleaseSemaphore(semaphoreI, 1, NULL);
   }
   return 0;
}//thread1


DWORD WINAPI thread2(LPVOID threadParam)
{
   while(runFlag)
   {
      WaitForSingleObject(semaphoreI, INFINITE);
      printf("%d ", count);
      ReleaseSemaphore(semaphoreP, 1, NULL);
   }
   return 0;
}//thread2
